import functools
from dataclasses import dataclass
from enum import Enum
from enum import auto
from typing import Any

from scim2_client import BaseSCIMClient
from scim2_client import SCIMClientError


class Status(Enum):
    SUCCESS = auto()
    ERROR = auto()


@dataclass
class CheckConfig:
    """Object used to configure the checks behavior."""

    client: BaseSCIMClient
    """The SCIM client that will be used to perform the requests."""

    raise_exceptions: bool = False
    """Whether to raise exceptions or store them in a :class:`~scim2_tester.Result` object."""


class SCIMTesterError(Exception):
    """Exception raised when a check failed and the `raise_exceptions` config parameter is :data:`True`."""

    def __init__(self, message: str, conf: CheckConfig):
        super().__init__()
        self.message = message
        self.conf = conf


@dataclass
class CheckResult:
    """Store a check result."""

    conf: CheckConfig
    status: Status

    title: str | None = None
    """The title of the check."""

    description: str | None = None
    """What the check does, and why the spec advises it to do."""

    reason: str | None = None
    """Why it failed, or how it succeed."""

    data: Any | None = None
    """Any related data that can help to debug."""

    def __post_init__(self):
        if self.conf.raise_exceptions and self.status == Status.ERROR:
            raise SCIMTesterError(self.title, self)


def checker(func):
    """Decorate checker methods.

    - It adds a title and a description to the returned result, extracted from the method name and its docstring.
    - It catches SCIMClient errors.
    """

    @functools.wraps(func)
    def wrapped(conf: CheckConfig, *args, **kwargs):
        try:
            result = func(conf, *args, **kwargs)
        except SCIMClientError as exc:
            if conf.raise_exceptions:
                raise
            result = CheckResult(
                conf, status=Status.ERROR, reason=str(exc), data=exc.source
            )

        # decorate results
        if isinstance(result, CheckResult):
            result.title = func.__name__
            result.description = func.__doc__
        else:
            result[0].title = func.__name__
            result[0].description = func.__doc__
        return result

    return wrapped



import re

from httpx import Client
from scim2_client.engines.httpx import SyncSCIMClient
from scim2_models import Error
from scim2_models import Group
from scim2_models import User

from scim2_tester.checker import check_random_url
from scim2_tester.utils import CheckConfig
from scim2_tester.utils import Status


def test_random_url(httpserver):
    """Test reaching a random URL that returns a SCIM 404 error."""
    httpserver.expect_request(re.compile(r".*")).respond_with_json(
        Error(status=404, detail="Endpoint Not Found").model_dump(),
        status=404,
        content_type="application/scim+json",
    )

    client = Client(base_url=f"http://localhost:{httpserver.port}")
    scim = SyncSCIMClient(client, resource_models=(User, Group))
    conf = CheckConfig(scim)
    result = check_random_url(conf)

    assert result.status == Status.SUCCESS
    assert "correctly returned a 404 error" in result.reason


def test_random_url_valid_object(httpserver):
    """Test reaching a random URL that returns a SCIM object."""
    httpserver.expect_request(re.compile(r".*")).respond_with_json(
        User(
            id="2819c223-7f76-453a-919d-413861904646", user_name="bjensen@example.com"
        ).model_dump(),
        status=404,
        content_type="application/scim+json",
    )

    client = Client(base_url=f"http://localhost:{httpserver.port}")
    scim = SyncSCIMClient(client, resource_models=(User, Group))
    conf = CheckConfig(scim)
    result = check_random_url(conf)

    assert result.status == Status.ERROR
    assert "did not return an Error object" in result.reason


def test_random_url_not_404(httpserver):
    """Test reaching a random URL that returns a SCIM object."""
    httpserver.expect_request(re.compile(r".*")).respond_with_json(
        Error(status=200, detail="Endpoint Not Found").model_dump(),
        status=200,
        content_type="application/scim+json",
    )

    client = Client(base_url=f"http://localhost:{httpserver.port}")
    scim = SyncSCIMClient(client, resource_models=(User, Group))
    conf = CheckConfig(scim)
    result = check_random_url(conf)

    assert result.status == Status.ERROR
    assert "did return an object, but the status code is 200" in result.reason



from scim2_models import Resource
from scim2_models import ServiceProviderConfig

from .utils import CheckConfig
from .utils import CheckResult
from .utils import Status
from .utils import checker


@checker
def check_service_provider_config_endpoint(
    conf: CheckConfig,
) -> tuple[Resource, CheckResult]:
    """As described in RFC7644 §4 <rfc7644#section-4>`, `/ServiceProviderConfig` is a mandatory endpoint, and should only be accessible by GET.

    .. todo::

        Check thet POST/PUT/PATCH/DELETE methods on the endpoint
    """
    response = conf.client.query(ServiceProviderConfig)
    return CheckResult(conf, status=Status.SUCCESS, data=response)



import pytest
from httpx import Client
from scim2_client import SCIMClientError
from scim2_client.engines.httpx import SyncSCIMClient
from scim2_models import Group
from scim2_models import User

from scim2_tester.checker import check_server


def test_raise_exceptions():
    """Test that exceptions are raised instead of stored in a Result object when 'raise_exceptions' is True."""
    client = Client(base_url="https://invalid.test")
    scim = SyncSCIMClient(client, resource_models=(User, Group))
    with pytest.raises(SCIMClientError):
        check_server(scim, raise_exceptions=True)



import random
import uuid
from enum import Enum
from inspect import isclass
from typing import get_origin

from pydantic import EmailStr
from scim2_models import ComplexAttribute
from scim2_models import EnterpriseUser
from scim2_models import Extension
from scim2_models import Group
from scim2_models import Meta
from scim2_models import Reference
from scim2_models import Resource
from scim2_models import ResourceType
from scim2_models import ServiceProviderConfig
from scim2_models import User

from scim2_tester.utils import CheckConfig
from scim2_tester.utils import CheckResult
from scim2_tester.utils import Status
from scim2_tester.utils import checker


def model_from_resource_type(conf: CheckConfig, resource_type: ResourceType):
    if resource_type.id == "User":
        return (
            User[EnterpriseUser]
            if User[EnterpriseUser] in conf.client.resource_models
            else User
        )

    if resource_type.id == "Group":
        return Group


def fill_with_random_values(obj) -> Resource:
    for field_name, field in obj.model_fields.items():
        if field.default:
            continue

        field_type = obj.get_field_root_type(field_name)
        is_multiple = obj.get_field_multiplicity(field_name)

        if field_type is Meta:
            value = None

        elif field_type is int:
            value = uuid.uuid4().int

        elif field_type is bool:
            value = random.choice([True, False])

        elif get_origin(field_type) is Reference:
            value = f"https://{str(uuid.uuid4())}.test"

        elif field_type is EmailStr:
            # pydantic won't allow to use the 'test' TLD here
            value = f"{uuid.uuid4()}@{uuid.uuid4()}.com"

        elif isclass(field_type) and issubclass(field_type, Enum):
            value = random.choice(list(field_type))

        elif isclass(field_type) and issubclass(field_type, ComplexAttribute):
            value = field_type()
            fill_with_random_values(value)

        elif isclass(field_type) and issubclass(field_type, Extension):
            value = field_type()
            fill_with_random_values(value)

        else:
            value = str(uuid.uuid4())

        if is_multiple:
            setattr(obj, field_name, [value])

        else:
            setattr(obj, field_name, value)

    return obj


@checker
def check_object_creation(
    conf: CheckConfig, obj: Resource
) -> tuple[Resource, CheckResult]:
    """Perform an object creation.

    Todo:
      - check if the fields of the result object are the same than the
      fields of the request object

    """
    response = conf.client.create(obj)

    return CheckResult(
        conf,
        status=Status.SUCCESS,
        reason=f"Successful creation of a {obj.__class__.__name__} object with id {response.id}",
        data=response,
    )


@checker
def check_object_query(
    conf: CheckConfig, obj: Resource
) -> tuple[Resource, CheckResult]:
    """Perform an object query by knowing its id.

    Todo:
      - check if the fields of the result object are the same than the
      fields of the request object

    """
    response = conf.client.query(obj.__class__, obj.id)
    return CheckResult(
        conf,
        status=Status.SUCCESS,
        reason=f"Successful query of a {obj.__class__.__name__} object with id {response.id}",
        data=response,
    )


@checker
def check_object_query_without_id(
    conf: CheckConfig, obj: Resource
) -> tuple[Resource, CheckResult]:
    """Perform an object creation.

    Todo:
      - look for the object across several pages
      - check if the fields of the result object are the same than the
      fields of the request object

    """
    response = conf.client.query(obj.__class__)
    found = any(obj.id == resource.id for resource in response.resources)
    if not found:
        return CheckResult(
            conf,
            status=Status.ERROR,
            reason=f"Could not find object {obj.__class__.__name__} with id : {response.detail}",
            data=response,
        )

    return CheckResult(
        conf,
        status=Status.SUCCESS,
        reason=f"Successful query of a {obj.__class__.__name__} object with id {obj.id}",
        data=response,
    )


@checker
def check_object_replacement(
    conf: CheckConfig, obj: Resource
) -> tuple[Resource, CheckResult]:
    """Perform an object replacement.

    Todo:
      - check if the fields of the result object are the same than the
      fields of the request object

    """
    response = conf.client.replace(obj)
    return CheckResult(
        conf,
        status=Status.SUCCESS,
        reason=f"Successful replacement of a {obj.__class__.__name__} object with id {response.id}",
        data=response,
    )


@checker
def check_object_deletion(
    conf: CheckConfig, obj: Resource
) -> tuple[Resource, CheckResult]:
    """Perform an object deletion."""
    conf.client.delete(obj.__class__, obj.id)
    return CheckResult(
        conf,
        status=Status.SUCCESS,
        reason=f"Successful deletion of a {obj.__class__.__name__} object with id {obj.id}",
    )


def check_resource_type(
    conf: CheckConfig,
    resource_type: ResourceType,
    service_provider_config: ServiceProviderConfig,
) -> list[CheckResult]:
    results = []

    model = model_from_resource_type(conf, resource_type)
    obj = model()
    fill_with_random_values(obj)

    result = check_object_creation(conf, obj)
    results.append(result)

    if result.status == Status.SUCCESS:
        created_obj = result.data
        result = check_object_query(conf, created_obj)
        queried_obj = result.data
        results.append(result)

        result = check_object_query_without_id(conf, created_obj)
        results.append(result)

        fill_with_random_values(queried_obj)
        result = check_object_replacement(conf, created_obj)
        results.append(result)

        result = check_object_deletion(conf, created_obj)
        results.append(result)

    return results



from enum import Enum

from scim2_models import ComplexAttribute
from scim2_models import Reference
from scim2_models import Resource

from scim2_tester.resource import fill_with_random_values


def test_random_values():
    """Check that 'fill_with_random_values' produce valid objects."""

    class Complex(ComplexAttribute):
        str_unique: str | None = None
        str_multiple: list[str] | None = None

    class CustomModel(Resource):
        schemas: list[str] = ["org:test:CustomModel"]

        class Type(str, Enum):
            foo = "foo"
            bar = "bar"

        str_unique: str | None = None
        str_multiple: list[str] | None = None

        int_unique: int | None = None
        int_multiple: list[int] | None = None

        bool_unique: bool | None = None
        bool_multiple: list[bool] | None = None

        reference_unique: Reference | None = None
        reference_multiple: list[Reference] | None = None

        email_unique: Reference | None = None
        email_multiple: list[Reference] | None = None

        enum_unique: Type | None = None
        enum_multiple: list[Type] | None = None

        complex_unique: Complex | None = None
        complex_multiple: list[Complex] | None = None

    obj = CustomModel()
    fill_with_random_values(obj)
    for field_name in obj.model_fields:
        if field_name == "meta":
            continue

        assert getattr(obj, field_name) is not None



import datetime
import os
import sys
from importlib import metadata

sys.path.insert(0, os.path.abspath(".."))
sys.path.insert(0, os.path.abspath("../scim2_tester"))

# -- General configuration ------------------------------------------------

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosectionlabel",
    "sphinx.ext.doctest",
    "sphinx.ext.graphviz",
    "sphinx.ext.intersphinx",
    "sphinx.ext.todo",
    "sphinx.ext.viewcode",
    "sphinx_click",
    "sphinx_issues",
    "sphinx_paramlinks",
    "sphinxcontrib.autodoc_pydantic",
    "myst_parser",
]

templates_path = ["_templates"]
master_doc = "index"
project = "scim2-tester"
year = datetime.datetime.now().strftime("%Y")
copyright = f"{year}, Yaal Coop"
author = "Yaal Coop"
source_suffix = {
    ".rst": "restructuredtext",
    ".txt": "markdown",
    ".md": "markdown",
}

version = metadata.version("scim2_tester")
language = "en"
exclude_patterns = []
pygments_style = "sphinx"
todo_include_todos = True
toctree_collapse = False

intersphinx_mapping = {
    "python": ("https://docs.python.org/3", None),
    "scim2_models": ("https://scim2-models.readthedocs.io/en/latest/", None),
    "scim2_client": ("https://scim2-client.readthedocs.io/en/latest/", None),
    "scim2_cli": ("https://scim2-cli.readthedocs.io/en/latest/", None),
    "werkzeug": ("https://werkzeug.palletsprojects.com", None),
}

# -- Options for HTML output ----------------------------------------------

html_theme = "shibuya"
# html_static_path = ["_static"]
html_baseurl = "https://scim2-tester.readthedocs.io"
html_theme_options = {
    "globaltoc_expand_depth": 3,
    "accent_color": "tomato",
    "github_url": "https://github.com/python-scim/scim2-tester",
    "mastodon_url": "https://toot.aquilenet.fr/@yaal",
    "nav_links": [
        {"title": "scim2-models", "url": "https://scim2-models.readthedocs.io"},
        {
            "title": "scim2-client",
            "url": "https://scim2-client.readthedocs.io",
        },
        {
            "title": "scim2-cli",
            "url": "https://scim2-cli.readthedocs.io",
        },
    ],
}

# -- Options for sphinx-issues -------------------------------------

issues_github_path = "python-scim/scim2-models"



from .checker import check_server
from .utils import CheckConfig
from .utils import CheckResult
from .utils import SCIMTesterError
from .utils import Status

__all__ = ["check_server", "Status", "CheckResult", "CheckConfig", "SCIMTesterError"]



import pytest
from scim2_client.engines.werkzeug import TestSCIMClient
from scim2_models import EnterpriseUser
from scim2_models import Group
from scim2_models import User
from scim2_server.backend import InMemoryBackend
from scim2_server.provider import SCIMProvider
from scim2_server.utils import load_default_resource_types
from scim2_server.utils import load_default_schemas

from scim2_tester import check_server

TestSCIMClient.__test__ = False


@pytest.fixture
def scim2_server():
    backend = InMemoryBackend()
    app = SCIMProvider(backend)

    for schema in load_default_schemas().values():
        app.register_schema(schema)

    for resource_type in load_default_resource_types().values():
        app.register_resource_type(resource_type)

    return app


def test_scim2_server(scim2_server):
    scim = TestSCIMClient(scim2_server, resource_models=(User[EnterpriseUser], Group))
    check_server(scim, raise_exceptions=True)



import re

from httpx import Client
from scim2_client.engines.httpx import SyncSCIMClient
from scim2_models import Context
from scim2_models import Error
from scim2_models import Group
from scim2_models import User

from scim2_tester.checker import check_schemas_endpoint
from scim2_tester.checker import check_server
from scim2_tester.resource import check_object_query
from scim2_tester.utils import CheckConfig
from scim2_tester.utils import Status


def test_unreachable_host():
    """Test reaching a invalid URL."""
    client = Client(base_url="https://invalid.test")
    scim = SyncSCIMClient(client, resource_models=(User, Group))
    results = check_server(scim)

    assert all(result.status == Status.ERROR for result in results)
    assert all(
        result.reason == "Network error happened during request" for result in results
    )


def test_bad_authentication(httpserver):
    """Test reaching a valid URL with incorrect authentication."""
    httpserver.expect_request(re.compile(r".*")).respond_with_json(
        Error(status=401, detail="Authentication is needed").model_dump(),
        status=401,
        content_type="application/scim+json",
    )

    client = Client(base_url=f"http://localhost:{httpserver.port}")
    scim = SyncSCIMClient(client, resource_models=(User, Group))
    conf = CheckConfig(scim)
    result = check_schemas_endpoint(conf)

    assert result.status == Status.ERROR
    assert (
        result.reason
        == "The server returned a SCIM Error object: Authentication is needed"
    )


def test_bad_content_type(httpserver):
    """Test reaching a valid URL returning an invalid content type."""
    scim_user = User(id="scim", username="scim")
    json_user = User(id="json", username="json")
    invalid_user = User(id="invalid", username="invalid")
    missing_user = User(id="missing", username="missing")

    httpserver.expect_request(re.compile(r"/Users/scim")).respond_with_json(
        scim_user.model_dump(scim_ctx=Context.RESOURCE_QUERY_RESPONSE),
        content_type="application/scim+json",
    )
    httpserver.expect_request(re.compile(r"/Users/json")).respond_with_json(
        json_user.model_dump(scim_ctx=Context.RESOURCE_QUERY_RESPONSE),
        content_type="application/json",
    )
    httpserver.expect_request(re.compile(r"/Users/invalid")).respond_with_json(
        invalid_user.model_dump(scim_ctx=Context.RESOURCE_QUERY_RESPONSE),
        content_type="application/invalid",
    )
    httpserver.expect_request(re.compile(r"/Users/missing")).respond_with_json(
        invalid_user.model_dump(scim_ctx=Context.RESOURCE_QUERY_RESPONSE),
        content_type="",
    )

    client = Client(base_url=f"http://localhost:{httpserver.port}")
    scim = SyncSCIMClient(client, resource_models=(User, Group))
    conf = CheckConfig(scim)

    result = check_object_query(conf, scim_user)
    assert result.status == Status.SUCCESS

    result = check_object_query(conf, json_user)
    assert result.status == Status.SUCCESS

    result = check_object_query(conf, invalid_user)
    assert result.status == Status.ERROR
    assert result.reason == "Unexpected content type: application/invalid"

    result = check_object_query(conf, missing_user)
    assert result.status == Status.ERROR
    assert result.reason == "Unexpected content type: "
